"""Implementation of schema builder for object type."""
from uuid import uuid4
from marshmallow import fields
from nadia.schema import NadiaSchema
from nadia.common import Builder


class ObjectBuilder(Builder):
    """Schema builder for object datatype."""

    def build_schema(self, spec, **kwargs):
        """Build Schema from a definition of OpenAPI object.

        :param spec: a mapping containing object definition extracted from OpenAPI spec.
        :type spec: dict
        :return: Schema constructed from `spec`.
        :rtype: :py:class:`marshmallow.Schema`
        """
        attrs = self.construct_attributes_schemas(spec)
        obj_schema = self.create_schema_type(attrs)
        return fields.Nested(obj_schema, **self.translate_args(spec, **kwargs))

    def construct_attributes_schemas(self, spec):
        """Construct Schemas corresponding to object definition.

        This method is used to construct attributes of the object schema being constructed.

        :param spec: an OpenAPI object's definition.
        :type spec: dict
        :return: a mapping property-name -> Schema or Field.
        :rtype: dict
        """
        properties = spec['properties']
        attr_schemas = {}
        for prop_name, prop_content in properties.items():
            attr_type = prop_content['type']
            attr_schemas_builder = self.builder_provider.get_builder(attr_type)
            attr_required = prop_name in spec.get('required', [])
            attr_schemas[prop_name] = attr_schemas_builder.build_schema(prop_content, required=attr_required)
        return attr_schemas

    @staticmethod
    def create_schema_type(attrs):
        """Create schema type from given dictionary of attributes.

        :param attrs: mapping of attributes of the newly created Python type.
        :type attrs: dict
        :return: newly created type with randomly chosen name and single base class -
         :py:class:`nadia.NadiaSchema`.
        :rtype: type
        """
        return type('Object' + str(uuid4()), (NadiaSchema,), attrs)



"""Common Exceptions."""


class UnknownTypeException(Exception):
    """Exception raised when we encounter unknown datatype in schema.

    Note that circumstances under which this Exception is raised should not
    happen when the schema dict passed was created from valid OpenAPI specification.

    :param typename: name of type that resulted in raising exception.
    :type typename: str
    """
    def __init__(self, typename):
        super(UnknownTypeException, self).__init__(typename)
        self.typename = typename

    def __str__(self):
        return 'Unknown data type: {}'.format(self.typename)



"""Base schema used by Nadia."""
from marshmallow import Schema, validates_schema, ValidationError


class NadiaSchema(Schema):
    """Base marshmallow schema used by Nadia."""

    @validates_schema(pass_original=True)
    def check_unknown_fields(self, _data, original_data):
        """Validator raising ValidationError if extra keys are provided.

        This validator is taken from the the official recipe located at:
        http://marshmallow.readthedocs.io/en/latest/extending.html#validating-original-input-data
        """
        unknown = set(original_data) - set(self.fields)
        if unknown:
            raise ValidationError('Unknown field', unknown)



"""Test cases for schema generator for object type."""
import unittest
from unittest import mock
from ddt import ddt, data
from marshmallow import fields, Schema
from nadia.object import ObjectBuilder
from nadia.common import Builder
from nadia.builder_provider import BuilderProvider
from nadia import primitives


@ddt
class TestObjectBuilder(unittest.TestCase):
    """Test case for ObjectBuilder."""

    provider = None
    properties = None
    required = None

    def setUp(self):
        number_mock = mock.Mock()
        number_mock.build_schema.return_value = fields.Float()
        int_mock = mock.Mock()
        int_mock.build_schema.return_value = fields.Int()
        str_mock = mock.Mock()
        str_mock.build_schema.return_value = fields.Str()
        self.provider = BuilderProvider({'number': number_mock, 'integer': int_mock,
                                         'string': str_mock})
        self.properties = {'x': {'type': 'number'},
                           'y': {'type': 'integer'},
                           'name': {'type': 'string'}}
        self.required = ['x', 'y']

    def test_return_type(self):
        """Building schema with ObjectBuilder should return marshmallow.fields.Nested instance."""
        provider = BuilderProvider({'number': mock.Mock(spec=Builder)})
        obj_def = {'type': 'object', 'properties': {'num1': {'type': 'number'}}}
        builder = ObjectBuilder(provider)
        self.assertEqual(fields.Nested, type(builder.build_schema(obj_def)))

    def test_nullable(self):
        """Building schema for object type should correctly use nullable field."""
        obj_def = {'type': 'object', 'properties': self.properties}
        builder = ObjectBuilder(self.provider)
        for nullable in (True, False):
            obj_def['nullable'] = nullable
            self.assertEqual(nullable, builder.build_schema(obj_def).allow_none)

    def test_required(self):
        """Building schema for object type should correctly use required field."""
        obj_def = {'type': 'object', 'properties': self.properties}
        builder = ObjectBuilder(self.provider)
        for required in (True, False):
            self.assertEqual(required, builder.build_schema(obj_def, required=required).required)

    def test_properties_building(self):
        """Constructing nested objects attributes should correctly call provided schema builders."""
        obj_def = {'type': 'object', 'properties': self.properties}
        builder = ObjectBuilder(self.provider)
        props_schemas = builder.construct_attributes_schemas(obj_def)
        for prop_name, prop in self.properties.items():
            prop_builder = self.provider.get_builder(prop['type'])
            prop_builder.build_schema.assert_called_once()
            self.assertEqual(props_schemas[prop_name], prop_builder.build_schema.return_value)

    @mock.patch('nadia.object.fields.Nested')
    @mock.patch('nadia.object.ObjectBuilder.create_schema_type')
    @mock.patch('nadia.object.ObjectBuilder.construct_attributes_schemas')
    def test_build_schema(self, cas, cst, nested):
        """Constructing object schema should result in correct call chain."""
        obj_def = {'type': 'object', 'properties': self.properties}
        builder = ObjectBuilder(self.provider)
        schema = builder.build_schema(obj_def)
        cas.assert_called_once_with(obj_def)
        cst.assert_called_once_with(cas.return_value)
        nested.assert_called_once_with(cst.return_value, allow_none=False, required=False)



"""Schema generators for primitive types."""
from marshmallow import fields
from nadia.common import Builder


class PrimitiveBuilder(Builder):
    """Base class for primitive fields builder."""
    key = None
    marshmallow_class = None

    @classmethod
    def build_schema(cls, spec, **kwargs):
        """Build a Field for a primitive object.

        .. seealso: :py:meth:`nadia.common.Builder.build_schema`

        .. note :: Conforming to the base class documentation, this method returns
           instances of :py:class:`marshmallow.Field`.
        """
        return cls.marshmallow_class(**cls.translate_args(spec, **kwargs))

class FloatBuilder(PrimitiveBuilder):
    """Float schema builder.

    This builder is designed for constructing schemas for OpenAPI `number` type.
    """
    key = 'number'
    marshmallow_class = fields.Float


class IntegerBuilder(PrimitiveBuilder):
    """Integer schema builder.

    This builder is designed for constructing schemas for OpenAPI 'integer' type.
    """
    key = 'integer'
    marshmallow_class = fields.Integer


class StringBuilder(PrimitiveBuilder):
    """Str schema builder.

    This builder is designed for constructing schemas for OpenAPI 'string' type.
    """
    key = 'string'
    marshmallow_class = fields.String



"""Test cases for builder_provider module."""
import unittest
from unittest import mock
from ddt import ddt, data
from nadia.builder_provider import BuilderProvider


@ddt
class TestBuilderProvider(unittest.TestCase):
    """Test case for BuilderProvider."""

    builders_dict = {'type1': mock.Mock(),
                     'type2': mock.Mock()}

    @data('type1', 'type2')
    def test_instantiates_correct_class(self, typename):
        """BuilderProvider should instantiate correct classes."""
        self.builders_dict[typename].reset_mock()
        provider = BuilderProvider(self.builders_dict)

        ret_val = provider.get_builder(typename)
        self.assertEqual(ret_val, self.builders_dict[typename].return_value)
        self.builders_dict[typename].assert_called_once_with(provider)

    def test_instantiates_builders_once(self):
        """BuilderProvider should cache created builder objects."""
        provider = BuilderProvider(self.builders_dict)
        provider.get_builder('type1')
        provider.get_builder('type1')
        self.builders_dict['type1'].assert_called_once_with(provider)



"""Test commoon functionallities for field builders."""
import unittest
from ddt import ddt, data
from nadia import primitives


@ddt
class TestPrimitiveBuilder(unittest.TestCase):
    """Test casefor functionallities common to all primitive builder classes."""

    @data(primitives.FloatBuilder,
          primitives.IntegerBuilder,
          primitives.StringBuilder)
    def test_default_args(self, builder):
        """FloatBuilder should set default values as in OpenAPI spec."""
        field = builder.build_schema({})
        self.assertEqual(False, field.allow_none)
        self.assertEqual(False, field.required)
        self.assertIsNone(field.validate)

    @data(primitives.FloatBuilder,
          primitives.IntegerBuilder,
          primitives.StringBuilder)
    def test_nullable(self, builder):
        """Building a Float field should respect nullable settings."""
        for nullable in (True, False):
            field = builder.build_schema({'nullable': nullable})
            self.assertEqual(nullable, field.allow_none)

    @data(primitives.FloatBuilder,
          primitives.IntegerBuilder,
          primitives.StringBuilder)
    def test_required(self, builder):
        """Building a Float field should respect required settings."""
        for required in (True, False):
            field = builder.build_schema({}, required=required)
            self.assertEqual(required, field.required)



"""Test case for schema generator for array type."""
import unittest
from unittest import mock
from ddt import ddt, data
from marshmallow import fields
from nadia.array import ArrayBuilder
from nadia.builder_provider import BuilderProvider
from nadia import primitives


@ddt
class TestArrayBuilder(unittest.TestCase):
    """Test case for ArrayBuilder."""

    def test_return_type(self):
        """Building schema with ArrayBuilder should return marshmallow.fields.List instance."""
        provider = BuilderProvider({'number': primitives.FloatBuilder})
        arr_def = {'type': 'array', 'items': {'type': 'number'}}
        builder = ArrayBuilder(provider)
        self.assertEqual(fields.List, type(builder.build_schema(arr_def)))

    @data(primitives.FloatBuilder, primitives.IntegerBuilder, primitives.StringBuilder)
    def test_primitives_array(self, primitive_builder):
        """Building schema for array of primitives should return List with correct container."""
        provider = BuilderProvider({primitive_builder.key: primitive_builder})
        arr_def = {'type': 'array', 'items': {'type': primitive_builder.key}}
        builder = ArrayBuilder(provider)
        schema = builder.build_schema(arr_def)
        self.assertEqual(type(schema.container), primitive_builder.marshmallow_class)

    def test_nullable(self):
        """Building schema for array type should correctly use nullable field."""
        provider = BuilderProvider({'number': primitives.FloatBuilder})
        arr_def = {'type': 'array', 'items': {'type': 'number'}}
        builder = ArrayBuilder(provider)
        for nullable in (True, False):
            arr_def['nullable'] = nullable
            schema = builder.build_schema(arr_def)
            self.assertEqual(nullable, schema.allow_none)

    def test_required(self):
        """Building schema for array type should correctly utilize rquired field."""
        provider = BuilderProvider({'number': primitives.FloatBuilder})
        arr_def = {'type': 'array', 'items': {'type': 'number'}}
        builder = ArrayBuilder(provider)
        for required in (True, False):
            schema = builder.build_schema(arr_def, required=required)
            self.assertEqual(required, schema.required)



"""Implementation of class responsible for obtaining schema builders for OpenAPI types."""
import logging
from functools import lru_cache
from nadia.array import ArrayBuilder
from nadia.object import ObjectBuilder
from nadia.primitives import FloatBuilder, IntegerBuilder, StringBuilder
from nadia.exceptions import UnknownTypeException


class BuilderProvider(object):
    """Class for providing builders for various data types.

    :param builders: mapping between OpenAPI types and classes implementing builder
     interface for them.
    :type builders: dict

    .. note:: The purpose of this class is to act as a mapping between types and
       corresponding builders, while elegantly handling unknown types and lazy creation
       of the builders. It is injected as a dependency in :py:class:`nadia.api.SchemaBuilder`.
    """
    def __init__(self, builders):
        self.builders = builders

    @lru_cache(maxsize=None)
    def get_builder(self, typename):
        """Get builder instance for given OpenAPI type.

        :param typename: OpenAPI type for which to get a builder.
        :type typename: str
        :return: a builder corresponding to given type.
        :rtype: a subclass of :py:class:`nadia.common.Builder`
        :raises: :py:exc:`nadia.exceptions.UnknownTypeException` if typename does not
         correspond to known OpenAPI type.
        """
        if typename not in self.builders:
            raise UnknownTypeException(typename)
        self.logger.info('Constructing builder for type: %s.', typename)
        return self.builders[typename](self)

    @property
    def logger(self):
        """Logger used by this BuilderProvider."""
        return logging.getLogger('BuilderProvider')

    @staticmethod
    def get_default():
        """Construct BuilderProvider with default builders.

        :rtype: :py:class:`nadia.builder_provider.BuilderProvider`
        """
        return BuilderProvider({
            'number': FloatBuilder,
            'integer': IntegerBuilder,
            'string': StringBuilder,
            'object': ObjectBuilder,
            'array': ArrayBuilder})



"""Functionalities related to building schemas for array type."""
from marshmallow import fields
from nadia.common import Builder


class ArrayBuilder(Builder):
    """Schema builder for array datatype."""

    def build_schema(self, spec, **kwargs):
        """Build Marshmallow schema for array datatype.

        :param spec: specification of array for which Schema should be build.
        :type spec: dict
        :returns: a List field constructed such that its properties correspond to the ones
         defined in `spec`.
        :rtype: :py:class:`marshmallow.fields.List`
        """
        item_body = spec['items']
        item_builder = self.builder_provider.get_builder(item_body['type'])
        return fields.List(item_builder.build_schema(item_body), **self.translate_args(spec, **kwargs))



"""The nadia public api."""
from uuid import uuid4
from nadia.builder_provider import BuilderProvider
from nadia.schema import NadiaSchema


class SchemaBuilder(object):
    """Class used for building schemas from given specification dict.

    :param builder_provider: an object providing builders via get_builder method.
     Typically an instance of :py:class:`nadia.builder_provider.BuilderProvider`.
    """

    def __init__(self, builder_provider):
        self.builder_provider = builder_provider

    def build(self, spec):
        """Build schema from given specification dict.

        :param spec: a dictionary containing specification of the object for which
         Schema should be build.
        :type spec: dict
        :return: Schema corresponding to the object defined by spec
        :rtype: :py:class:`marshmallow.Schema`
        """
        content_builder = self.builder_provider.get_builder(spec.get('type', 'object'))
        attrs = {'content':  content_builder.build_schema(spec)}
        return type('Object' + str(uuid4()), (NadiaSchema, ), attrs)()

    @staticmethod
    def create():
        """Create SchemaBuilder.

        :rtype: :py:class:`nadia.api.SchemaBuilder`

        .. note:: This method is designed to be further extended as a factory method.
           Later it should be able to accept some parameters governing creation of the
           SchemaBuilder. Currently it creates :py:class:`nadia.api.SchemaBuilder`
           by passing default instance of :py:class:`nadia.builder_provider.BulderProvider`
           to initializer.
        """
        provider = BuilderProvider.get_default()
        return SchemaBuilder(provider)



"""Functionalities related to all builders."""
from abc import ABC, abstractmethod


class Builder(ABC):
    """Base class for all field builders.

    :param builder_provider: a provider used for obtaining builders for other OpenAPI types.
    :type bulder_provider: a subclass of :py:class:`nadia.builder_provider.BuilderProvider`
    """
    key = None
    marshmallow_class = None

    def __init__(self, builder_provider):
        self.builder_provider = builder_provider

    @staticmethod
    def translate_args(spec, **kwargs):
        """Translate arguments given in OpenAPI spec to keyword arguments for marshmallow classes.

        :param spec: a dictionary extracted from OpenAPI spec containing definition of some
         object.
        :type spec: dict
        :return: A mapping containing keyword arguments used for constructing marshmallow objects.
        :rtype: dict
        """
        
        return {
            'allow_none': spec.get('nullable', False),
            'required': kwargs.get('required', False)
        }

    @abstractmethod
    def build_schema(self, spec, **kwargs):
        """Build marshmallow schema from definition extracted from OpenAPI specs.

        :param spec: an object definition extracted from OpenAPI specification.
        :return: Schema or a Field constructed from the spec dictionary.
        :rtype: :py:class:`marshmallow.Schema` or :py:class:marshmallow.Field

        .. note:: The return type as well as the nature of the object constructed depends
           on the concrete implementation of Builder. Usually the builder designed for
           handling object types will return Schema object, while builders designed for handling
           other data types will return Field instances.
        """
        pass



"""Nadia: automatic generation of Marshmallow Schemas for OpenAPI 3 schemas."""

from setuptools import setup, find_packages

CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.6',
    'Topic :: Software Development']

with open('README.md') as readme:
    LONG_DESCRIPTION = readme.read()

setup(
    name='nadia',
    description=__doc__,
    use_scm_version=True,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    platforms=["Linux", "Unix"],
    setup_requires=['setuptools_scm'],
    install_requires=['marshmallow', 'pyaml'],
    tests_require=['marshmallow', 'pyaml', 'ddt'],
    author='Konrad Jałowiecki <dexter2206@gmail.com>, Magdalena Nowak',
    author_email='dexter2206@gmail.com',
    packages=find_packages(exclude=["tests", "tests.*"]),
    keywords='openapi schema validation'
)



"""Test cases for schema generators for primitive types."""
import unittest
from marshmallow import fields
from nadia import primitives


class TestFloatBulder(unittest.TestCase):
    """Test case for FloatBuilder."""

    def test_correct_type(self):
        """Building a Float field return marshmallow.fields.Float."""
        field = primitives.FloatBuilder.build_schema({})
        self.assertEqual(fields.Float, type(field))
